# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Helpers for working with models in CLI scripts."""
from __future__ import annotations

__all__ = [
    "ModelCLI",
    "ModelInfoCLI",
    "enumerate_known_models",
    "loadcli",
    "loadinfo",
]

import collections.abc as cabc
import importlib.abc
import importlib.resources as imr
import json
import logging
import os
import typing as t

import capellambse

LOGGER = logging.getLogger(__name__)

try:
    import click

    class ModelCLI(click.ParamType):
        """Declare an option that loads a model.

        Use instances of this class for the *type* argument to
        :func:`click.option` etc.

        See Also
        --------
        capellambse.cli_helpers.loadcli :
            A standalone function performing the same task.

        Examples
        --------
        .. code-block:: python

           @click.command()
           @click.option("-m", "--model", type=capellambse.ModelCLI())
           def main(model: capellambse.MelodyModel) -> None:
               ...
        """

        name = "CAPELLA_MODEL"

        def convert(self, value: t.Any, param, ctx) -> capellambse.MelodyModel:
            if isinstance(value, capellambse.MelodyModel):
                return value

            try:
                return loadcli(value)
            except ValueError as err:
                self.fail(err.args[0], param, ctx)
                assert False

    class ModelInfoCLI(click.ParamType):
        """Declare an option that loads information about a model.

        Use instances of this class for the *type* argument to
        :func:`click.option` etc.

        See Also
        --------
        capellambse.cli_helpers.loadinfo :
            A standalone function performing the same task.

        Examples
        --------
        .. code-block:: python

           @click.command()
           @click.open("-m", "modelinfo", type=capellambse.ModelInfoCLI())
           def main(modelinfo: dict[str, t.Any]) -> None:
               # change any options, for example:
               modelinfo["diagram_cache"] = "/tmp/diagrams"
               model = capellambse.MelodyModel(**modelinfo)
        """

        name = "CAPELLA_MODEL"

        def convert(self, value: t.Any, param, ctx) -> dict[str, t.Any]:
            if isinstance(value, dict):
                return value

            try:
                return loadinfo(value)
            except ValueError as err:
                self.fail(err.args[0], param, ctx)
                assert False

except ImportError:
    if not t.TYPE_CHECKING:

        def ModelCLI(*__, **_):
            """Raise a dependency error."""
            raise RuntimeError("click is not installed")

        def ModelInfoCLI(*__, **_):
            """Raise a dependency error."""
            raise RuntimeError("click is not installed")


def enumerate_known_models() -> cabc.Iterator[importlib.abc.Traversable]:
    """Enumerate the models that are found in the ``known_models`` folders.

    Two places are searched for models: The *known_models* folder in the
    user's configuration directory, and the *known_models* folder in the
    installed ``capellambse`` package.

    Run the following command to print the location of the user's
    *known_models* folder:

    .. code:: bash

       python -m capellambse.cli_helpers

    In order to make a custom model known, place a JSON file in one of
    these *known_models* folders. It should contain a dictionary with
    the keyword arguments to :class:`~capellambse.model.MelodyModel` -
    specifically it needs a ``path``, optionally an ``entrypoint``, and
    any additional arguments that the underlying
    :class:`~capellambse.filehandler.FileHandler` might need to gain
    access to the model.

    Files in the user's configuration directory take precedence over
    files in the package directory. If a file with the same name exists
    in both places, the one in the user's configuration directory will
    be used.

    Be aware that relative paths in the JSON will be interpreted
    relative to the current working directory.
    """
    names = set[str]()
    userdir = capellambse.dirs.user_config_path.joinpath("known_models")
    configdir = imr.files(capellambse).joinpath("known_models")
    i: importlib.abc.Traversable
    try:
        for i in userdir.iterdir():
            if i.name.endswith(".json") and i.is_file():
                names.add(i.name)
                yield i
    except FileNotFoundError:
        pass
    for i in configdir.iterdir():
        if i.name.endswith(".json") and i.is_file() and i.name not in names:
            yield i


def loadcli(value: str | os.PathLike[str]) -> capellambse.MelodyModel:
    """Load a model from a file or JSON string.

    This function works like :func:`loadinfo`, and also loads the model
    for convenience.

    Parameters
    ----------
    value
        As described for :func:`loadinfo`.

    Returns
    -------
    ~capellambse.model.MelodyModel
        The loaded model, as described by the *value*.

    Examples
    --------
    .. code-block:: python

       def main():
           model = capellambse.loadcli(sys.argv[1])
    """
    modelinfo = loadinfo(value)
    LOGGER.info("Loading model from %s", modelinfo["path"])
    return capellambse.MelodyModel(**modelinfo)


def loadinfo(value: str | os.PathLike[str]) -> dict[str, t.Any]:
    """Load information about how to load a model as dict.

    Parameters
    ----------
    value
        One of the following:

        - A str or PathLike pointing to an ``.aird`` file
        - A str or PathLike pointing to a ``.json`` file, which
          contains the arguments to instantiate a
          :class:`~capellambse.model.MelodyModel`
        - The contents of such a JSON file (as string)

    Returns
    -------
    dict[str, ~typing.Any]
        A dict with information about how to load a
        :class:`~capellambse.model.MelodyModel`.

    Raises
    ------
    TypeError
        If the *value* cannot be parsed as described above.
    ValueError
        If the *value* looks like a "known model" name, but the name is
        not defined.

    Examples
    --------
    .. code-block:: python

       def main():
           modelinfo = capellambse.loadinfo(sys.argv[1])
           # change any options, for example:
           modelinfo["diagram_cache"] = "/tmp/diagrams"
           model = MelodyModel(**modelinfo)
    """
    if isinstance(value, str):
        if value.endswith((".aird", ".json")):
            return _load_from_file(value)
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return _load_from_file(value)

    if isinstance(value, os.PathLike):
        value = os.fspath(value)
        if isinstance(value, str):
            return _load_from_file(value)

    raise TypeError("value must be a str or PathLike returning str")


def _load_from_file(value: str) -> dict[str, t.Any]:
    if value.endswith(".aird"):
        return {"path": value}

    if value.endswith(".json"):
        with open(value, encoding="utf-8") as file:
            return json.load(file)

    name = value + ".json"
    for known in enumerate_known_models():
        if known.name == name:
            return json.loads(known.read_text(encoding="utf-8"))

    if os.path.exists(value):
        return {"path": value}

    raise ValueError(
        "value is not a known model nor contains valid JSON"
    ) from None


def _main() -> None:
    print("User-defined 'known models' are searched in:")
    print()
    print(f"   {capellambse.dirs.user_config_path / 'known_models'}")
    print()
    print("The following models are currently known:")
    for file in enumerate_known_models():
        name, _, _ = file.name.rpartition(".")
        print(f"  - {name}")


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import warnings

warnings.warn(
    "{__name__} is deprecated, use capellambse.diagram instead",
    DeprecationWarning,
    stacklevel=2,
)
del warnings

# pylint: disable=unused-wildcard-import, wildcard-import
from capellambse.diagram import *
from capellambse.diagram._diagram import __all__



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Factory functions for Edges inside a diagram."""
from __future__ import annotations

import collections.abc as cabc
import dataclasses
import math
import typing as t

from capellambse import diagram, helpers

from . import _common as C
from . import _filters, _styling

# Port classes that should not be considered
# when differentiating ``PortAllocation``s.
PORTALLOCATION_CLASS_BLACKLIST = {
    "CP",
    "CP_IN",
    "CP_OUT",
    "CP_INOUT",
    "CP_UNSET",
}
XT_EXITEM = "org.polarsys.capella.core.data.information:ExchangeItem"


def generic_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create an Edge from the diagram XML."""
    bendpoints, sourceport, targetport = extract_bendpoints(seb)

    try:
        ostyle = next(seb.diag_element.iterchildren("ownedStyle"))
    except StopIteration:
        raise ValueError(
            "Cannot find style definition for edge"
            f" {seb.data_element.attrib['element']}"
        ) from None
    routingstyle = ostyle.attrib.get("routingStyle")

    if not bendpoints:
        if routingstyle == "manhattan":
            bendpoints = route_manhattan(sourceport, targetport)
        elif routingstyle == "tree":
            bendpoints = route_tree(sourceport, targetport)
        else:
            bendpoints = route_oblique(sourceport, targetport)

    edge = diagram.Edge(
        bendpoints,
        source=sourceport,
        target=targetport,
        uuid=seb.data_element.attrib["element"],
        styleclass=seb.styleclass,
    )

    _filters.setfilters(seb, edge)
    # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
    edge.styleoverrides = _styling.apply_style_overrides(
        seb.target_diagram.styleclass, f"Edge.{seb.styleclass}", ostyle
    )  # type: ignore[assignment]
    edge.labels.extend(_construct_labels(edge, seb))

    if isinstance(targetport, diagram.Box):
        snaptarget(edge, -1, -2, targetport, not edge.hidden, routingstyle)
    if isinstance(sourceport, diagram.Box):
        snaptarget(edge, 0, 1, sourceport, not edge.hidden, routingstyle)

    sourceport.add_context(seb.data_element.attrib["element"])
    if targetport is not None:
        targetport.add_context(seb.data_element.attrib["element"])

    return edge


def extract_bendpoints(
    seb: C.ElementBuilder,
) -> tuple[
    list[diagram.Vector2D], diagram.DiagramElement, diagram.DiagramElement
]:
    """Extract the bendpoints from an Edge's XML.

    Parameters
    ----------
    seb
        The element builder instance

    Returns
    -------
    tuple
        A 3-tuple with the list of bendpoints, the source element and
        the target element.
    """
    uid = (
        seb.data_element.attrib.get("element")
        or seb.data_element.attrib[C.ATT_XMID]
    )
    sourceport, targetport = get_end_ports(seb)
    sourcebounds = sourceport.bounds

    try:
        bendpoints_elm = next(seb.data_element.iterchildren("bendpoints"))
    except StopIteration:
        raise ValueError(f"No bendpoints definition for {uid!r}") from None

    bendpoints = []
    bendpoint_type = bendpoints_elm.attrib.get(C.ATT_XMT)

    if bendpoint_type == "notation:RelativeBendpoints":
        # Translate relative to absolute coordinates
        try:
            sourceanchor = next(
                seb.data_element.iterchildren("sourceAnchor")
            ).attrib["id"]
        except (StopIteration, KeyError):
            sourceanchor = "(0.5, 0.5)"
        if sourceanchor.endswith(" custom"):
            sourceanchor = sourceanchor[: -len(" custom")]
        sourceanchor = helpers.ssvparse(
            sourceanchor, float, parens="()", num=2
        )
        refpos = sourcebounds.pos + sourcebounds.size @ sourceanchor

        bendpoints_attrib = bendpoints_elm.attrib.get(
            "points", "[0, 0, 0, 0]$[0, 0, 0, 0]"
        )
        for point in bendpoints_attrib.split("$"):
            bendpoints.append(
                refpos + helpers.ssvparse(point, int, parens="[]", num=4)[:2]
            )

        if len(bendpoints) == 1 or all(b == bendpoints[0] for b in bendpoints):
            bendpoints = []
    else:
        raise ValueError(f"Unknown bendpoint type {bendpoint_type}")

    return bendpoints, sourceport, targetport


def get_end_ports(
    seb: C.ElementBuilder,
) -> tuple[diagram.DiagramElement, diagram.DiagramElement]:
    """Retrieve the source and target port of an Edge in the diagram."""

    def get_port_object(portside: str) -> diagram.DiagramElement:
        try:
            port = seb.data_element.attrib[portside]
            try:
                return seb.target_diagram[port]
            except KeyError:
                elem = seb.melodyloader.follow_link(seb.data_element, port)
                return seb.target_diagram[elem.attrib["element"]]
        except KeyError:
            uid = (
                seb.data_element.attrib.get("element")
                or seb.data_element.attrib[C.ATT_XMID]
            )
            C.LOGGER.warning(
                "Cannot draw edge %r: %s missing from diagram", uid, portside
            )
            raise C.SkipObject() from None

    return (get_port_object("source"), get_port_object("target"))


def route_manhattan(
    source: diagram.DiagramElement,
    target: diagram.DiagramElement,
) -> list[diagram.Vector2D]:
    """Reroute an Edge using the "Manhattan" routing style.

    Parameters
    ----------
    points
        The list of points
    source
        A Box to originate from
    target
        A Box or Edge to point to
    """

    source_point = source.vector_snap(
        source.center,
        source=target.center,
        style=diagram.RoutingStyle.MANHATTAN,
    )
    target_point = target.vector_snap(
        target.center,
        source=source.center,
        style=diagram.RoutingStyle.MANHATTAN,
    )

    if abs(source_point.x - target_point.x) > abs(
        source_point.y - target_point.y
    ):
        p1 = diagram.Vector2D(
            (source_point.x + target_point.x) / 2, source_point.y
        )
        p2 = diagram.Vector2D(p1.x, target_point.y)
    else:
        p1 = diagram.Vector2D(
            source_point.x, (source_point.y + target_point.y) / 2
        )
        p2 = diagram.Vector2D(target_point.x, p1.y)

    return [source_point, p1, p2, target_point]


def route_tree(
    source: diagram.DiagramElement,
    target: diagram.DiagramElement,
) -> list[diagram.Vector2D]:
    """Reroute an Edge using the "Tree" routing style.

    Parameters
    ----------
    source
        A Box to originate from
    target
        A Box or Edge to point to
    """
    sourcebounds = source.bounds
    sourcecenter = sourcebounds.pos + sourcebounds.size / 2
    targetbounds = target.bounds
    targetcenter = targetbounds.pos + targetbounds.size / 2

    source_y = sourcebounds.pos.y + sourcebounds.size.y * (
        sourcecenter.y > targetcenter.y
    )
    target_y = targetbounds.pos.y + targetbounds.size.y * (
        targetcenter.y > sourcecenter.y
    )
    centerpoint_y = (source_y + target_y) / 2

    return [
        diagram.Vector2D(sourcecenter.x, source_y),
        diagram.Vector2D(sourcecenter.x, centerpoint_y),
        diagram.Vector2D(targetcenter.x, centerpoint_y),
        diagram.Vector2D(targetcenter.x, target_y),
    ]


def route_oblique(
    source: diagram.DiagramElement,
    target: diagram.DiagramElement,
) -> list[diagram.Vector2D]:
    """Reroute an Edge using the "Oblique" routing style."""
    return [source.center, target.center]


def snaptarget(
    points: cabc.MutableSequence[diagram.Vector2D],
    i: int,
    next_i: int,
    target: diagram.DiagramElement,
    movetarget: bool = False,
    routingstyle: str | None = None,
) -> None:
    """Snap an Edge's end and (optionally) its target into place.

    Parameters
    ----------
    points
        A list of points describing the entire edge
    i
        The end point index in ``points``
    next_i
        The next point's index in ``points``
    target
        The Edge target (either a Box or another Edge)
    movetarget
        Allow to move the target under certain conditions
    routingstyle
        The routing style of the edge that caused this call
    """
    del movetarget

    if i < 0:
        i += len(points)
    if next_i < 0:
        next_i += len(points)
    assert i >= 0
    assert next_i >= 0

    if not diagram.SNAPPING or target is None:
        return

    if routingstyle == "manhattan":
        snap_manhattan(points, i, next_i, target)
    elif routingstyle == "tree":
        snap_tree(points, i, next_i, target)
    else:
        snap_oblique(points, i, next_i, target)


def snap_oblique(
    points: t.MutableSequence[diagram.Vector2D],
    i: int,
    next_i: int,
    target: diagram.DiagramElement,
) -> None:
    """Snap ``points``' end to ``target`` in a straight line."""
    points[i] = target.vector_snap(
        points[i], source=points[next_i], style=diagram.RoutingStyle.OBLIQUE
    )


def snap_manhattan(
    points: t.MutableSequence[diagram.Vector2D],
    i: int,
    next_i: int,
    target: diagram.DiagramElement,
) -> None:
    """Snap ``points``' end to ``target`` with axis-parallel lines."""
    direction = points[i] - points[next_i]
    axis = direction.closestaxis()
    manhattan = math.isclose(axis.angleto(direction), 0)
    if not manhattan:
        translate = points[next_i] @ (not axis.x, not axis.y)
        end = points[i] @ abs(axis)
        points[i] = end + translate

    endpoint = target.vector_snap(
        points[i], source=points[next_i], style=diagram.RoutingStyle.MANHATTAN
    )

    if axis.x:
        if math.isclose(endpoint.y, points[i].y):
            points[i] = endpoint
        else:
            points[i] = diagram.Vector2D(endpoint.x, points[i].y)
            points.insert(i + (i > next_i), endpoint)
    else:
        if math.isclose(endpoint.x, points[i].x):
            points[i] = endpoint
        else:
            points[i] = diagram.Vector2D(points[i].x, endpoint.y)
            points.insert(i + (i > next_i), endpoint)


def snap_tree(
    points: t.MutableSequence[diagram.Vector2D],
    i: int,
    next_i: int,
    target: diagram.DiagramElement,
) -> None:
    """Snap ``points``' end to ``target`` in tree routing style."""
    endpoint = target.vector_snap(
        points[i], source=points[next_i], style=diagram.RoutingStyle.TREE
    )

    if math.isclose(endpoint.x, points[i].x):
        points[i] = endpoint
    else:
        points[i] = diagram.Vector2D(endpoint.x, points[i].y)
        points.insert(
            i + (i > next_i), diagram.Vector2D(endpoint.x, points[next_i].y)
        )


def _construct_labels(
    edge: diagram.Edge, seb: C.SemanticElementBuilder
) -> list[diagram.Box]:
    """Construct the label box for an edge."""
    refpoints = _find_refpoints(edge)
    layouts = seb.data_element.xpath("./children/layoutConstraint")
    labels: list[diagram.Box] = []
    for (labelanchor, travel_direction), layout, melodyobj in zip(
        refpoints, layouts, seb.melodyobjs
    ):
        labeltext = melodyobj.get("name", "")

        label_pos = diagram.Vector2D(
            int(layout.get("x", "0")),
            int(layout.get("y", "0")),
        )
        label_size = diagram.Vector2D(
            int(layout.attrib.get("width", "0")),
            int(layout.attrib.get("height", "0")),
        )

        # Rotate the position vector into place
        label_pos = label_pos.rotatedby(travel_direction.angleto((1, 0)))

        labels.append(
            C.CenterAnchoredBox(
                labelanchor + label_pos,
                label_size,
                label=labeltext,
                styleclass="EdgeAnnotation",
            )
        )
    return labels


def _find_refpoints(
    points: cabc.Sequence[diagram.Vector2D],
) -> list[tuple[diagram.Vector2D, diagram.Vector2D]]:
    """Calculate the center point of the edge described by `points`."""
    refpoints: list[tuple[diagram.Vector2D, diagram.Vector2D]] = []

    lengths = [
        (points[i] - points[i + 1]).length for i in range(len(points) - 1)
    ]
    total_length = sum(lengths)
    refpoint_lengths = [
        total_length * 0.15,
        total_length * 0.5,
        total_length * 0.85,
    ]
    current_length = 0.0
    current_position = points[0]
    new_length = lengths[0]
    i = 0

    for next_refpoint in refpoint_lengths:
        while new_length < next_refpoint:
            i += 1
            current_length = new_length
            new_length += lengths[i]
            current_position = points[i]

        try:
            dir = (points[i + 1] - points[i]).normalized
        except ZeroDivisionError:  # pragma: no cover
            dir = diagram.Vector2D(1, 0)
        pos = current_position + dir * (next_refpoint - current_length)
        refpoints.append((pos, dir))

    refpoints[0], refpoints[1] = refpoints[1], refpoints[0]
    return refpoints


def labelless_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create an edge that should never have a label."""
    edge = generic_factory(seb)
    edge.labels = []
    return edge


def port_allocation_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Specially handle ``PortAllocation`` type edges.

    Generic PortAllocations need to be differentiated into specific port
    allocation types (e.g. FIPAllocation), based on the types of ports
    they connect with.
    """
    portclasses = set()

    for port in get_end_ports(seb):
        if (
            port.styleclass
            and port.styleclass not in PORTALLOCATION_CLASS_BLACKLIST
        ):
            portclasses.add(port.styleclass)

    if portclasses:
        seb.styleclass = f"{'_'.join(sorted(portclasses))}Allocation"
    return generic_factory(seb)


def state_transition_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create a StateTransition.

    StateTransitions (connecting two Modes or two States) do not use the
    transition's name as their label, but rather some transition-
    specific attributes.
    """
    edge = generic_factory(seb)
    if edge.labels:  # pragma: no branch
        triggers = seb.melodyloader.follow_links(
            seb.melodyobjs[0], seb.melodyobjs[0].get("triggers", "")
        )
        label = ", ".join(
            i.get("name", "(unnamed trigger)")
            for i in triggers
            if i is not None
        )

        if guard := _guard_condition(seb, "guard"):
            label = f"{label} [{guard}]"

        effects = seb.melodyloader.follow_links(
            seb.melodyobjs[0], seb.melodyobjs[0].get("effect", "")
        )
        if effects:
            effects_str = ", ".join(
                i.get("name", "") for i in effects if i is not None
            )
            label = f"{label} / {effects_str}"

        edge.labels[0].label = label
    return edge


def sequence_link_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create a SequenceLink.

    Sequence links (in Operational Process Diagrams) use the guard
    condition as label, if it is set. Otherwise the label stays empty.
    """
    edge = generic_factory(seb)
    guard = _guard_condition(seb, "condition")
    if guard and edge.labels:
        edge.labels[0].label = guard
    return edge


def constraint_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create the edge for a Constraint.

    As the Box already contains the label, it is not needed on the Edge.

    See Also
    --------
    capellambse.aird._box_factories.constraint_factory :
        The accompanying box factory.
    """
    edge = generic_factory(seb)
    edge.labels = []
    return edge


def fcil_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create a FunctionalChainInvolvementLinks."""
    seb.melodyobjs[0] = seb.melodyloader.follow_link(
        seb.melodyobjs[0], seb.melodyobjs[0].get("involved")
    )
    xtype = helpers.xtype_of(seb.melodyobjs[0])
    assert xtype is not None
    seb.styleclass = xtype.split(":")[-1]
    edge = generic_factory(seb)
    edge.labels = edge.labels[:1]
    assert edge.styleclass is not None and edge.target is not None
    if edge.target.styleclass == "OperationalActivity":
        edge.styleclass = "OperationalExchange"
    return edge


def eie_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create an exchange item element link."""
    edge = generic_factory(seb)
    edge.labels = edge.labels[:1]
    return edge


def fex_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create a functional exhcange."""
    edge = generic_factory(seb)
    assert edge.styleclass is not None and edge.target is not None
    if edge.target.styleclass == "OperationalActivity":
        edge.styleclass = "OperationalExchange"
    return edge


def _guard_condition(seb: C.SemanticElementBuilder, attr: str) -> str:
    """Extract the guard condition's text from the XML."""
    try:
        guard = seb.melodyloader.follow_links(
            seb.melodyobjs[0], seb.melodyobjs[0].get(attr, "")
        )[0]
    except IndexError:
        return ""
    else:
        return C.get_spec_text(
            dataclasses.replace(seb, melodyobjs=[guard, *seb.melodyobjs[1:]])
        )


def req_relation_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create a Capella Incoming or Outgoing Relation."""
    label = seb.melodyobjs[0].attrib.get("name", "")
    if not label:
        try:
            reltype_id = seb.melodyobjs[0].attrib["relationType"]
            reltype = seb.melodyloader[reltype_id]
            label = reltype.attrib["ReqIFLongName"]
        except KeyError:
            C.LOGGER.warning(
                "Requirement-Relation %r has no RelationType",
                seb.data_element.attrib[C.ATT_XMID],
            )
        finally:
            seb.melodyobjs[0].attrib["name"] = label

    return generic_factory(seb)


def include_extend_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create an AbstractCapabilityIncludes or -Extends edge."""
    if seb.melodyobjs[0].get("name") is None:
        seb.melodyobjs[0].attrib["name"] = seb.diag_element.get("name", "")
    return generic_factory(seb)


def association_factory(seb: C.SemanticElementBuilder) -> diagram.Edge:
    """Create an Association."""
    edge = generic_factory(seb)
    for member in seb.melodyobjs[0].iterchildren("ownedMembers"):
        kind = member.get("aggregationKind", "ASSOCIATION").capitalize()
        if kind != "Association":
            edge.styleclass = kind
    return edge



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implements the Capella Filtering extension."""
from __future__ import annotations

import collections.abc as cabc
import logging
import pathlib
import shutil
import sys
import typing as t

import typing_extensions as te

import capellambse
import capellambse.model.common as c
from capellambse import _native, helpers, model

VIEWPOINT: t.Final = "org.polarsys.capella.filtering"
NAMESPACE: t.Final = "http://www.polarsys.org/capella/filtering/6.0.0"
SYMBOLIC_NAME: t.Final = "filtering"

_LOGGER = logging.getLogger(__name__)

c.XTYPE_ANCHORS[__name__] = SYMBOLIC_NAME


@c.xtype_handler(None)
class FilteringCriterion(c.GenericElement):
    """A single filtering criterion."""

    _xmltag = "ownedFilteringCriteria"

    filtered_objects = c.ReferenceSearchingAccessor[c.GenericElement](
        (), "filtering_criteria", aslist=c.MixedElementList
    )


@c.xtype_handler(None)
class FilteringCriterionPkg(c.GenericElement):
    """A package containing multiple filtering criteria."""

    _xmltag = "ownedFilteringCriterionPkgs"

    criteria = c.DirectProxyAccessor(FilteringCriterion, aslist=c.ElementList)
    packages: c.Accessor[FilteringCriterionPkg]


@c.xtype_handler(None)
class FilteringModel(c.GenericElement):
    """A filtering model containing criteria to filter by."""

    criteria = c.DirectProxyAccessor(FilteringCriterion, aslist=c.ElementList)
    criterion_packages = c.DirectProxyAccessor(
        FilteringCriterionPkg, aslist=c.ElementList
    )


class AssociatedCriteriaAccessor(
    c.accessors.PhysicalAccessor[FilteringCriterion]
):
    def __init__(self) -> None:
        super().__init__(
            FilteringCriterion, aslist=c.ElementList[FilteringCriterion]
        )

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self: ...

    @t.overload
    def __get__(
        self, obj: c.ModelObject, objtype: type[t.Any] | None = ...
    ) -> c.ElementList[c.T]: ...

    def __get__(
        self,
        obj: c.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | c.ElementList[c.T]:
        del objtype
        if obj is None:  # pragma: no cover
            return self

        loader = obj._model._loader
        try:
            xt_critset = f"{SYMBOLIC_NAME}:AssociatedFilteringCriterionSet"
            critset = next(loader.iterchildren_xt(obj._element, xt_critset))
        except StopIteration:
            elems = []
        else:
            links = critset.get("filteringCriteria", "")
            elems = list(loader.follow_links(obj._element, links))

        rv = self._make_list(obj, elems)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv


@c.xtype_handler(None)
class FilteringResult(c.GenericElement):
    """A filtering result."""


@c.xtype_handler(None)
class ComposedFilteringResult(c.GenericElement):
    """A composed filtering result."""


def init() -> None:
    c.set_accessor(
        model.MelodyModel,
        "filtering_model",
        c.DirectProxyAccessor(FilteringModel, rootelem=model.XT_SYSENG),
    )
    c.set_accessor(
        c.GenericElement, "filtering_criteria", AssociatedCriteriaAccessor()
    )


c.set_self_references(
    (FilteringCriterionPkg, "packages"),
)

try:
    import click

    from capellambse import cli_helpers
except ImportError:

    def _main() -> None:
        """Display a dependency error."""
        print("Missing dependency 'capellambse[cli]'", file=sys.stderr)
        raise SystemExit(1)

else:

    @click.command()
    @click.argument("command")
    @click.option(
        "-m",
        "--model",
        "model_",
        type=cli_helpers.ModelCLI(),
        required=True,
        help="Source Capella model to derive from",
    )
    @click.option(
        "-r",
        "--result",
        "result_strings",
        multiple=True,
        help="A filtering result object, specified either by UUID or name.",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Path(
            file_okay=False, dir_okay=True, path_type=pathlib.Path
        ),
        default=".",
        help="Directory to store the derived project(s) in.",
    )
    @click.option(
        "-p",
        "--name-pattern",
        default="{model.name}_derived_{result.name}",
        help=(
            "The 'str.format()' pattern for generating output project names."
            " Available format arguments are 'model' and the current 'result'."
        ),
    )
    @click.option("--exe", help="Name or path of the Capella executable")
    @click.option("--docker", help="Name of a Docker image containing Capella")
    def _main(
        command: str,
        model_: capellambse.MelodyModel,
        result_strings: list[str],
        output: pathlib.Path,
        name_pattern: str,
        exe: str | None,
        docker: str | None,
    ) -> None:
        """Manage filtered Capella projects.

        Currently the only supported COMMAND is "derive", which performs
        project derivation based on an existing FilteringResult.

        The native Capella installation (options ``--exe`` or
        ``--docker``) may contain the placeholder '{VERSION}', which
        will be replaced by the Capella version number that the model
        was created with (e.g. "6.0.0").

        Exactly one of ``--exe`` or ``--docker`` must be specified.
        """
        if command != "derive":
            click.echo(f"Unsupported command: {command}", err=True)
            raise SystemExit(1)

        results = _find_results(model_, result_strings)
        output.mkdir(exist_ok=True, parents=True)
        with (
            _native.native_capella(model_, exe=exe, docker=docker) as cli,
            click.progressbar(
                results, item_show_func=lambda i: getattr(i, "name", "")
            ) as progress,
        ):
            assert cli.workspace
            for result in progress:
                derived_name = name_pattern.format(model=model_, result=result)
                output_sub = output / derived_name
                try:
                    shutil.rmtree(output_sub)
                except FileNotFoundError:
                    pass
                proc = cli(
                    *_native.ARGS_CMDLINE,
                    "-appid",
                    "org.polarsys.capella.filtering.commandline",
                    "-input",
                    "main_model",
                    "-filteringresultid",
                    result.uuid,
                    "-derivationprojectname",
                    derived_name,
                )
                derived = cli.workspace / derived_name
                if not derived.exists():
                    _LOGGER.error(
                        "Failed to derive result %r (%s)\n\n%s",
                        result.name,
                        result.uuid,
                        proc.stdout,
                    )
                    raise RuntimeError(
                        "Output directory does not exist for derived result"
                        f" {result.name!r} ({result.uuid})"
                    )
                shutil.copytree(derived, output_sub)

    def _find_results(
        loaded_model: capellambse.MelodyModel, result_strings: list[str]
    ) -> cabc.Sequence[FilteringResult | ComposedFilteringResult]:
        all_results: c.ElementList[FilteringResult | ComposedFilteringResult]
        all_results = loaded_model.search(
            FilteringResult, ComposedFilteringResult
        )
        if not result_strings:
            return all_results

        wanted: list[FilteringResult | ComposedFilteringResult] = []
        for result in result_strings:
            if helpers.is_uuid_string(result):
                try:
                    obj = loaded_model.by_uuid(result)
                except KeyError:
                    click.echo(
                        f"Error: No model object with UUID {result} found!",
                        err=True,
                    )
                    raise SystemExit(1) from None
                if not isinstance(
                    obj, (FilteringResult, ComposedFilteringResult)
                ):
                    click.echo(
                        f"Error: Object {result} is of type"
                        f" {type(obj).__name__}, expected a FilteringResult or"
                        f" ComposedFilteringResult",
                        err=True,
                    )
                    raise SystemExit(1)
                wanted.append(obj)
            else:
                objs = all_results.by_name(result, single=False)
                if len(objs) < 1:
                    click.echo(
                        f"Error: No result found with name: {result}", err=True
                    )
                    raise SystemExit(1)
                if len(objs) > 1:
                    click.echo(
                        f"Error: Ambiguous result name, found {len(objs)}"
                        f" objects: {result}",
                        err=True,
                    )
                    raise SystemExit(1)
                wanted.extend(objs)
        return wanted


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The decoration factories for svg elements."""
from __future__ import annotations

import collections.abc as cabc
import dataclasses
import logging
import re

logger = logging.getLogger(__name__)

icon_size = 20
"""Default icon size."""
icon_padding = 2
"""Default icon padding (left/right side)."""
feature_space = 24
"""Default margins/padding (top/bot and left/right) for feature text."""

function_ports = {"FIP", "FOP"}
directed_component_ports = {"CP_IN", "CP_OUT"}
component_ports = directed_component_ports | {"PP", "CP_INOUT", "CP_UNSET"}
all_ports = function_ports | component_ports
all_directed_ports = directed_component_ports | function_ports
start_aligned = {
    "LogicalComponent",
    "LogicalActor",
    "LogicalHumanActor",
    "LogicalHumanComponent",
}
only_icons = {"Requirement"}
DiagramClass = str
FaultyClass = str
PatchClass = str
needs_feature_line = {
    "BooleanType",
    "Class",
    "Enumeration",
    "NumericType",
    "PhysicalQuantity",
    "StringType",
}
always_top_label = needs_feature_line | {
    "Note",
    "RepresentationLink",
    "OperationalActivity",
    "PhysicalBehaviorComponent",
    "PhysicalBehaviorHumanComponent",
    "PhysicalBehaviorActor",
    "PhysicalBehaviorHumanActor",
    "PhysicalNodeComponent",
    "PhysicalNodeHumanComponent",
    "PhysicalNodeActor",
    "PhysicalNodeHumanActor",
}


@dataclasses.dataclass
class DecoFactory:
    function: cabc.Callable
    dependencies: tuple[str, ...]


class DecoFactories(dict[str, DecoFactory]):
    def __call__(
        self,
        func: cabc.Callable | None = None,
        dependencies: cabc.Iterable[str] = (),
    ) -> cabc.Callable:
        def decorator(func: cabc.Callable) -> cabc.Callable:
            symbol_name = re.sub(
                "(?:^|_)([a-z])",
                lambda m: m.group(1).capitalize(),
                func.__name__,
            )
            self[symbol_name] = DecoFactory(func, tuple(dependencies))
            return func

        if func is None:
            return decorator
        return decorator(func)

    def __missing__(self, class_: str) -> DecoFactory:
        logger.error("%s wasn't found in factories.", class_)
        assert "ErrorSymbol" in self
        return self["ErrorSymbol"]


deco_factories = DecoFactories()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import warnings

warnings.warn(
    f"{__name__} is deprecated, use capellambse.filehandler instead",
    DeprecationWarning,
    stacklevel=2,
)

del warnings

from capellambse.filehandler import (  # pylint: disable=unused-import
    FileHandler,
    TransactionClosedError,
    __all__,
    get_filehandler,
    load_entrypoint,
    split_protocol,
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Some additional glue to tie ReqIF into capellambse."""

__all__ = ["init"]

from capellambse.model import common as c
from capellambse.model import crosslayer

from . import _capellareq as cr
from . import _requirements as rq


def init() -> None:
    c.set_accessor(
        rq.Folder,
        "folders",
        c.DirectProxyAccessor(rq.Folder, aslist=c.ElementList),
    )
    c.set_accessor(
        c.GenericElement, "requirements", cr.ElementRelationAccessor()
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_modules",
        c.DirectProxyAccessor(cr.CapellaModule, aslist=c.ElementList),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirements",
        c.DeepProxyAccessor(
            rq.Requirement, aslist=c.ElementList, rootelem=cr.CapellaModule
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_types_folders",
        c.DirectProxyAccessor(cr.CapellaTypesFolder, aslist=c.ElementList),
    )
    c.set_accessor(
        cr.CapellaModule,
        "requirement_types_folders",
        c.DirectProxyAccessor(cr.CapellaTypesFolder, aslist=c.ElementList),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirement_types",
        c.DeepProxyAccessor(
            rq.RequirementType,
            aslist=c.ElementList,
            rootelem=cr.CapellaTypesFolder,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_module_types",
        c.DeepProxyAccessor(
            rq.ModuleType, aslist=c.ElementList, rootelem=cr.CapellaTypesFolder
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_relation_types",
        c.DeepProxyAccessor(
            rq.RelationType,
            aslist=c.ElementList,
            rootelem=cr.CapellaTypesFolder,
        ),
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "ModelObject",
    "GenericElement",
    "ElementList",
    "CachedElementList",
    "MixedElementList",
    "attr_equal",
]

import collections.abc as cabc
import enum
import functools
import inspect
import operator
import re
import textwrap
import typing as t

import markupsafe
import typing_extensions as te
from lxml import etree

import capellambse
from capellambse import helpers

from . import XTYPE_HANDLERS, T, U, accessors, properties

_NOT_SPECIFIED = object()
"Used to detect unspecified optional arguments"

_MapFunction: te.TypeAlias = (
    "cabc.Callable[[T], GenericElement | cabc.Iterable[GenericElement]]"
)


def attr_equal(attr: str) -> cabc.Callable[[type[T]], type[T]]:
    def add_wrapped_eq(cls: type[T]) -> type[T]:
        orig_eq = cls.__eq__
        orig_hash = cls.__hash__

        @functools.wraps(orig_eq)
        def new_eq(self, other: object) -> bool:
            # pylint: disable=unnecessary-dunder-call
            try:
                cmpkey = getattr(self, attr)
            except AttributeError:
                pass
            else:
                if isinstance(other, type(cmpkey)):
                    result = cmpkey.__eq__(other)
                else:
                    result = other.__eq__(cmpkey)
                if result is not NotImplemented:
                    return result
            return orig_eq(self, other)

        @functools.wraps(orig_hash)
        def new_hash(self) -> int:
            import warnings

            # <https://github.com/DSD-DBS/py-capellambse/issues/52>
            warnings.warn(
                (
                    "Hashing of this type is broken and will likely be"
                    " removed in the future. Please use the `.uuid` or"
                    f" `.{attr}` directly instead."
                ),
                category=FutureWarning,
                stacklevel=2,
            )

            return orig_hash(self)

        cls.__eq__ = new_eq  # type: ignore[method-assign]
        cls.__hash__ = new_hash  # type: ignore[method-assign]

        return cls

    return add_wrapped_eq


class ModelObject(t.Protocol):
    """A class that wraps a specific model object.

    Most of the time, you'll want to subclass the concrete
    ``GenericElement`` class. However, some special classes (e.g. AIRD
    diagrams) provide a compatible interface, but it doesn't make sense
    to wrap a specific XML element. This protocol class is used in type
    annotations to catch both "normal" GenericElement subclasses and the
    mentioned special cases.
    """

    @property
    def _model(self) -> capellambse.MelodyModel: ...

    @property
    def _element(self) -> etree._Element: ...

    @property
    def _constructed(self) -> bool: ...

    def __init__(
        self,
        model: capellambse.MelodyModel,
        parent: etree._Element,
        xmltag: str | None,
        /,
        **kw: t.Any,
    ) -> None:
        """Create a new model object.

        Parameters
        ----------
        model
            The model instance.
        parent
            The parent XML element below which to create a new object.
        kw
            Any additional arguments will be used to populate the
            instance attributes. Note that some attributes may be
            required by specific element types at construction time
            (commonly e.g. ``uuid``).
        """

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: t.Any
    ) -> ModelObject:
        """Instantiate a ModelObject from existing model elements."""


class GenericElement:
    """Provides high-level access to a single model element."""

    uuid = properties.AttributeProperty("id", writable=False)
    xtype = property(lambda self: helpers.xtype_of(self._element))
    name = properties.AttributeProperty("name", optional=True, default="")
    description = properties.HTMLAttributeProperty(
        "description", optional=True
    )
    summary = properties.AttributeProperty("summary", optional=True)
    diagrams: accessors.Accessor[capellambse.model.diagram.Diagram]
    diagrams = property(  # type: ignore[assignment]
        lambda self: self._model.diagrams.by_target_uuid(self.uuid)
    )

    constraints: accessors.Accessor
    parent: accessors.ParentAccessor

    _constructed: bool
    _required_attrs = frozenset({"uuid", "xtype"})
    _xmltag: str | None = None

    @property
    def progress_status(self) -> properties.AttributeProperty | str:
        uuid = self._element.get("status")
        if uuid is None:
            return "NOT_SET"

        return self.from_model(self._model, self._model._loader[uuid]).name

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: etree._Element
    ) -> te.Self:
        """Wrap an existing model object.

        Parameters
        ----------
        model
            The MelodyModel instance
        element
            The XML element to wrap

        Returns
        -------
        obj
            An instance of GenericElement (or a more appropriate
            subclass, if any) that wraps the given XML element.
        """
        class_ = cls
        if class_ is GenericElement:
            xtype = helpers.xtype_of(element)
            if xtype is not None:
                ancestors = model._loader.iterancestors(element)
                for ancestor in ancestors:
                    anc_xtype = helpers.xtype_of(ancestor)
                    try:
                        class_ = XTYPE_HANDLERS[anc_xtype][xtype]
                    except KeyError:
                        pass
                    else:
                        break
                else:
                    try:
                        class_ = XTYPE_HANDLERS[None][xtype]
                    except KeyError:
                        pass
            if class_ is not cls:
                return class_.from_model(model, element)
        self = class_.__new__(class_)
        self._model = model
        self._element = element
        self._constructed = True
        return self

    def __init__(
        self,
        model: capellambse.MelodyModel,
        parent: etree._Element,
        xmltag: str | None = None,
        /,
        **kw: t.Any,
    ) -> None:
        all_required_attrs: set[str] = set()
        for basecls in type(self).mro():
            all_required_attrs |= getattr(
                basecls, "_required_attrs", frozenset()
            )
        missing_attrs = all_required_attrs - frozenset(kw)
        if missing_attrs:
            mattrs = ", ".join(sorted(missing_attrs))
            raise TypeError(f"Missing required keyword arguments: {mattrs}")

        super().__init__()
        if xmltag is None:
            xmltag = self._xmltag
        if xmltag is None:
            raise TypeError(
                f"Cannot instantiate {type(self).__name__} directly"
            )
        self._constructed = False
        self._model = model
        self._element: etree._Element = etree.Element(xmltag)
        parent.append(self._element)
        try:
            for key, val in kw.items():
                if key == "xtype":
                    self._element.set(helpers.ATT_XT, val)
                    self._model._loader.add_namespace(
                        parent, val.split(":", maxsplit=1)[0]
                    )
                elif not isinstance(
                    getattr(type(self), key),
                    (accessors.Accessor, properties.AttributeProperty),
                ):
                    raise TypeError(
                        f"Cannot set {key!r} on {type(self).__name__}"
                    )
                else:
                    setattr(self, key, val)
            self._model._loader.idcache_index(self._element)
        except BaseException:
            parent.remove(self._element)
            raise
        self._constructed = True

    def __setattr__(self, attr: str, value: t.Any) -> None:
        if attr.startswith("_") or hasattr(type(self), attr):
            super().__setattr__(attr, value)
        else:
            raise AttributeError(
                f"{attr!r} isn't defined on {type(self).__name__}"
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._element is other._element

    def __hash__(self):
        return hash(self._element)

    def __repr__(self) -> str:  # pragma: no cover
        header = self._short_repr_()

        attrs: list[str] = []
        for attr in dir(self):
            if attr.startswith("_"):
                continue

            acc = getattr(type(self), attr, None)
            if isinstance(acc, accessors.ReferenceSearchingAccessor):
                classes = ", ".join(i.__name__ for i in acc.target_classes)
                attrs.append(
                    f".{attr} = ... # backreference to {classes}"
                    " - omitted: can be slow to compute"
                )
                continue

            try:
                value = getattr(self, attr)
            except Exception:
                continue

            if inspect.ismethod(value):
                continue

            if hasattr(value, "_short_repr_"):
                value_repr = f"{value._short_repr_()}"
            else:
                value_repr = textwrap.shorten(repr(value), 250)

            prefix = f".{attr} = "
            blankprefix = " " * len(prefix)
            value_repr = "\n".join(
                (prefix, blankprefix)[bool(i)] + line
                for i, line in enumerate(value_repr.splitlines() or [""])
            )
            attrs.append(value_repr)

        attr_text = "\n".join(attrs)
        return f"{header}\n{attr_text}"

    def _short_repr_(self) -> str:
        # pylint: disable=unidiomatic-typecheck
        if type(self) is GenericElement:
            mytype = f"Model element ({self.xtype})"
        else:
            mytype = type(self).__name__
        if self.name:
            name = f" {self.name!r}"
        else:
            name = ""
        return f"<{mytype}{name} ({self.uuid})>"

    def __html__(self) -> markupsafe.Markup:
        fragments: list[str] = []
        escape = markupsafe.Markup.escape

        # pylint: disable=unidiomatic-typecheck
        if type(self) is GenericElement:
            fragments.append("<h1>Model element")
        else:
            fragments.append("<h1>")
            fragments.append(escape(self.name or type(self).__name__))
        fragments.append(' <span style="font-size: 70%;">(')
        fragments.append(escape(self.xtype))
        fragments.append(")</span>")
        fragments.append("</h1>")

        fragments.append("<table>")
        for attr in dir(self):
            if attr.startswith("_"):
                continue

            acc = getattr(type(self), attr, None)
            if isinstance(acc, accessors.ReferenceSearchingAccessor):
                classes = ", ".join(i.__name__ for i in acc.target_classes)
                fragments.append('<tr><th style="text-align: right;">')
                fragments.append(escape(attr))
                fragments.append('</th><td style="text-align: left;"><em>')
                fragments.append(f"Backreference to {escape(classes)}")
                fragments.append(" - omitted: can be slow to compute.")
                fragments.append(" Display this property directly to show.")
                fragments.append("</em></td></tr>")
                continue

            try:
                value = getattr(self, attr)
            except Exception:
                continue

            if inspect.ismethod(value):
                continue

            fragments.append('<tr><th style="text-align: right;">')
            fragments.append(escape(attr))
            fragments.append('</th><td style="text-align: left;">')

            if hasattr(value, "_short_html_"):
                fragments.append(value._short_html_())
            elif isinstance(value, str):
                fragments.append(escape(value))
            else:
                value = repr(value)
                if len(value) > 250:
                    value = value[:250] + " [...]"
                fragments.append("<em>")
                fragments.append(escape(value))
                fragments.append("</em>")
            fragments.append("</td></tr>")
        fragments.append("</table>")
        return markupsafe.Markup("".join(fragments))

    def _short_html_(self) -> markupsafe.Markup:
        return self._wrap_short_html(
            f" &quot;{markupsafe.Markup.escape(self.name)}&quot;"
            f"{(': ' + str(self.value)) if hasattr(self, 'value') else ''}"
        )

    def _wrap_short_html(self, content: str) -> markupsafe.Markup:
        return markupsafe.Markup(
            f"<strong>{markupsafe.Markup.escape(type(self).__name__)}</strong>"
            f"{content} ({markupsafe.Markup.escape(self.uuid)})"
        )

    def _repr_html_(self) -> str:
        return self.__html__()

    if t.TYPE_CHECKING:

        def __getattr__(self, attr: str) -> t.Any:
            """Account for extension attributes in static type checks."""


class ElementList(cabc.MutableSequence, t.Generic[T]):
    """Provides access to elements without affecting the underlying model."""

    __slots__ = (
        "_elemclass",
        "_elements",
        "_ElementList__mapkey",
        "_ElementList__mapvalue",
        "_model",
    )

    class _Filter(t.Generic[U]):
        """Filters this list based on an extractor function."""

        __slots__ = ("_attr", "_parent", "_positive", "_single")

        def __init__(
            self,
            parent: ElementList[T],
            attr: str,
            *,
            positive: bool = True,
            single: bool = False,
        ) -> None:
            """Create a filter object.

            If the extractor returns an :class:`enum.Enum` member for any
            element, the enum member's name will be used for comparisons
            against any ``values``.

            Parameters
            ----------
            parent
                Reference to the :class:`ElementList` this filter should
                operate on
            attr
                The attribute on list members to filter on
            extract_key
                Callable that extracts the key from an element
            positive
                Use elements that match (True) or don't (False)
            single
                When listing all matches, return a single element
                instead. If multiple elements match, it is an error; if
                none match, a ``KeyError`` is raised. Can be overridden
                at call time.
            """
            self._attr = attr
            self._parent = parent
            self._positive = positive
            self._single = single

        def extract_key(self, element: T) -> U | str:
            extractor = operator.attrgetter(self._attr)
            value: U | enum.Enum | str = extractor(element)
            if isinstance(value, enum.Enum):
                value = value.name
            return value

        def make_values_container(self, *values: U) -> cabc.Container[U]:
            return values

        def ismatch(self, element: T, valueset: cabc.Container[U]) -> bool:
            try:
                value = self.extract_key(element)
            except AttributeError:
                return False

            return self._positive == (value in valueset)

        def __call__(
            self, *values: U, single: bool | None = None
        ) -> T | ElementList[T]:
            """List all elements that match this filter.

            Parameters
            ----------
            values
                The values to match against.
            single
                If not ``None``, overrides the ``single`` argument to
                the constructor for this filter call.
            """
            if single is None:
                single = self._single
            valueset = self.make_values_container(*values)
            indices = []
            elements = []
            for i, elm in enumerate(self._parent):
                if self.ismatch(elm, valueset):
                    indices.append(i)
                    elements.append(self._parent._elements[i])

            if not single:
                return self._parent._newlist(elements)
            if len(elements) > 1:
                value = values[0] if len(values) == 1 else values
                raise KeyError(f"Multiple matches for {value!r}")
            if len(elements) == 0:
                raise KeyError(values[0] if len(values) == 1 else values)
            return self._parent[indices[0]]  # Ensure proper construction

        def __iter__(self) -> cabc.Iterator[U | str]:
            """Yield values that result in a non-empty list when filtered for.

            The returned iterator yields all values that, when given to
            :meth:`__call__`, will result in a non-empty list being
            returned. Consequently, if the original list was empty, this
            iterator will yield no values.

            The order in which the values are yielded is undefined.
            """
            # Use list, since not all elements may be hashable.
            yielded: set[U | str] = set()

            for elm in self._parent:
                key = self.extract_key(elm)
                if key not in yielded:
                    yield key
                    yielded.add(key)

        def __contains__(self, value: U) -> bool:
            valueset = self.make_values_container(value)
            for elm in self._parent:
                if self.ismatch(elm, valueset):
                    return True
            return False

        def __getattr__(self, attr: str) -> ElementList._Filter[U]:
            if attr.startswith("_"):
                raise AttributeError(f"Invalid filter attribute name: {attr}")
            return type(self)(
                self._parent,
                f"{self._attr}.{attr}",
                positive=self._positive,
                single=self._single,
            )

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[T] | None = None,
        *,
        mapkey: str | None = None,
        mapvalue: str | None = None,
    ) -> None:
        self._model = model
        self._elements = elements
        if elemclass is not None:
            self._elemclass = elemclass
        else:
            self._elemclass = GenericElement  # type: ignore[assignment]

        if bool(mapkey) != bool(mapvalue):
            raise TypeError(
                "mapkey and mapvalue must both either be set or unset"
            )
        # pylint: disable=assigning-non-slot  # false-positive
        if not mapkey or not mapvalue:
            self.__mapkey: cabc.Callable[[T], t.Any] | None = None
            self.__mapvalue: str | None = None
        else:
            self.__mapkey = operator.attrgetter(mapkey)
            self.__mapvalue = mapvalue

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, cabc.Sequence):
            return NotImplemented
        return len(self) == len(other) and all(
            ours == theirs for ours, theirs in zip(self, other)
        )

    def __add(
        self, other: object, *, reflected: bool = False
    ) -> ElementList[T]:
        if not isinstance(other, ElementList):
            return NotImplemented
        if self._model is not other._model:
            raise ValueError("Cannot add ElementLists from different models")

        if self._elemclass is other._elemclass is not GenericElement:
            listclass: type[ElementList] = ElementList
            elemclass: type[ModelObject] = self._elemclass
        else:
            listclass = MixedElementList
            elemclass = GenericElement

        if not reflected:
            elements = self._elements + other._elements
        else:
            elements = other._elements + self._elements

        return listclass(self._model, elements, elemclass)

    def __add__(self, other: object) -> ElementList[T]:
        return self.__add(other)

    def __radd__(self, other: object) -> ElementList[T]:
        return self.__add(other, reflected=True)

    def __sub(
        self, other: object, *, reflected: bool = False
    ) -> ElementList[T]:
        if not isinstance(other, cabc.Sequence):
            return NotImplemented

        if reflected:
            if isinstance(other, ElementList):
                objclass = other._elemclass
            else:
                objclass = GenericElement
        else:
            objclass = self._elemclass

        base: cabc.Sequence[t.Any]
        if not reflected:
            base = self
            excluded = set(i.uuid for i in other)
        else:
            base = other
            excluded = set(i.uuid for i in self)

        return ElementList(
            self._model,
            [i._element for i in base if i.uuid not in excluded],
            objclass,
        )

    def __sub__(self, other: object) -> ElementList[T]:
        """Return a new list without elements found in ``other``."""
        return self.__sub(other)

    def __rsub__(self, other: object) -> ElementList[T]:
        return self.__sub(other, reflected=True)

    def __len__(self) -> int:
        return len(self._elements)

    @t.overload
    def __getitem__(self, idx: int) -> T: ...

    @t.overload
    def __getitem__(self, idx: slice) -> ElementList[T]: ...

    @t.overload
    def __getitem__(self, idx: str) -> t.Any: ...

    def __getitem__(self, idx: int | slice | str) -> t.Any:
        if isinstance(idx, slice):
            return self._newlist(self._elements[idx])
        if isinstance(idx, str):
            obj = self._map_find(idx)
            return self._map_getvalue(obj)
        return self._elemclass.from_model(self._model, self._elements[idx])

    @t.overload
    def __setitem__(self, index: int, value: T) -> None: ...

    @t.overload
    def __setitem__(self, index: slice, value: cabc.Iterable[T]) -> None: ...

    @t.overload
    def __setitem__(self, index: str, value: t.Any) -> None: ...

    def __setitem__(self, index: int | slice | str, value: t.Any) -> None:
        if isinstance(index, slice):
            del self[index]
            for i, element in enumerate(value, start=index.start):
                self.insert(i, element)
        elif isinstance(index, int):
            del self[index]
            self.insert(index, value)
        else:
            obj = self._map_find(index)
            self._map_setvalue(obj, value)

    def __delitem__(self, index: int | slice) -> None:
        del self._elements[index]

    def __getattr__(
        self,
        attr: str,
    ) -> cabc.Callable[..., T | ElementList[T]]:
        if attr.startswith("by_"):
            attr = attr[len("by_") :]
            if attr in {"name", "uuid"}:
                return self._Filter(self, attr, single=True)
            return self._Filter(self, attr)

        if attr.startswith("exclude_") and attr.endswith("s"):
            attr = attr[len("exclude_") : -len("s")]
            return self._Filter(self, attr, positive=False)

        return getattr(super(), attr)

    def __dir__(self) -> list[str]:  # pragma: no cover
        no_dir_attr = re.compile(r"^(_|as_|pvmt$|nodes$|diagrams?$)")

        def filterable_attrs() -> cabc.Iterator[str]:
            for obj in self:
                try:
                    obj_attrs = dir(obj)
                except Exception:
                    pass
                for attr in obj_attrs:
                    if not no_dir_attr.search(attr) and isinstance(
                        getattr(obj, attr), str
                    ):
                        yield f"by_{attr}"
                        yield f"exclude_{attr}s"

        attrs = list(super().__dir__())
        attrs.extend(filterable_attrs())
        return attrs

    def __repr__(self) -> str:  # pragma: no cover
        if not self:
            return "[]"

        items: list[str] = []
        for i, item in enumerate(self):
            if hasattr(item, "_short_repr_"):
                rv = item._short_repr_()
            else:
                rv = repr(item)
            r = rv.splitlines() or [""]
            prefix = f"[{i}] "
            r[0] = prefix + r[0]
            prefix = " " * len(prefix)
            r[1:] = [prefix + l for l in r[1:]]
            items.append("\n".join(r))
        return "\n".join(items)

    def _short_repr_(self) -> str:
        return repr(self)

    def __html__(self) -> markupsafe.Markup:
        if not self:
            return markupsafe.Markup("<p><em>(Empty list)</em></p>")

        fragments = ['<ol start="0" style="text-align: left;">']
        for i in self:
            fragments.append(f"<li>{i._short_html_()}</li>")
        fragments.append("</ol>")
        return markupsafe.Markup("".join(fragments))

    def _short_html_(self) -> markupsafe.Markup:
        return self.__html__()

    def _repr_html_(self) -> str:
        return self.__html__()

    def _mapkey(self, obj: T) -> t.Any:
        if self.__mapkey is None or self.__mapvalue is None:
            raise TypeError("This list cannot act as a mapping")

        try:
            return self.__mapkey(obj)
        except AttributeError:
            return None

    def _map_find(self, key: str) -> T:
        """Find the target object of a mapping operation.

        When this list acts as a mapping (like ``some_list["key"]``),
        this method finds the target object associated with the
        ``"key"``.

        See Also
        --------
        :meth:`_map_getvalue` and :meth:`_map_setvalue`
            Get or set the mapping value behind the target object.
        """
        if self.__mapkey is None or self.__mapvalue is None:
            raise TypeError("This list cannot act as a mapping")

        candidates = [i for i in self if self.__mapkey(i) == key]
        if len(candidates) > 1:
            raise ValueError(f"Multiple matches for key {key!r}")
        if not candidates:
            raise KeyError(key)
        return candidates[0]

    def _map_getvalue(self, obj: T) -> t.Any:
        """Get the mapping value from the target object."""
        assert self.__mapvalue
        getvalue = operator.attrgetter(self.__mapvalue)
        return getvalue(obj)

    def _map_setvalue(self, obj: T, value: t.Any) -> None:
        """Set a new mapping value on the target object."""
        assert self.__mapvalue
        key = self.__mapvalue.rsplit(".", maxsplit=1)
        if len(key) == 1:
            target: t.Any = obj
        else:
            target = operator.attrgetter(key[0])(obj)

        setattr(target, key[-1], value)

    def _newlist(self, elements: list[etree._Element]) -> ElementList[T]:
        listtype = self._newlist_type()
        return listtype(self._model, elements, self._elemclass)

    def _newlist_type(self) -> type[ElementList]:
        return type(self)

    @t.overload
    def get(self, key: str) -> T | None: ...

    @t.overload
    def get(self, key: str, default: U) -> T | U: ...

    def get(self, key: str, default: t.Any = None) -> t.Any:
        try:
            return self[key]
        except KeyError:
            return default

    def insert(self, index: int, value: T) -> None:
        elm: etree._Element = value._element
        self._elements.insert(index, elm)

    def items(self) -> ElementListMapItemsView[T]:
        return ElementListMapItemsView(self)

    def keys(self) -> ElementListMapKeyView:
        return ElementListMapKeyView(self)

    def values(self) -> ElementList[T]:
        return self

    def filter(
        self, predicate: str | cabc.Callable[[T], bool]
    ) -> ElementList[T]:
        """Filter this list with a custom predicate.

        The predicate may be the name of an attribute or a callable,
        which will be called on each list item. If the attribute value
        or the callable's return value is truthy, the item is included
        in the resulting list.

        When specifying the name of an attribute, nested attributes can
        be chained using ``.``, like ``"parent.name"`` (which would
        pick all elements whose ``parent`` has a non-empty ``name``).
        """
        if isinstance(predicate, str):
            predicate = operator.attrgetter(predicate)
        return self._newlist([i._element for i in self if predicate(i)])

    def map(self, attr: str | _MapFunction[T]) -> ElementList[GenericElement]:
        """Apply a function to each element in this list.

        If the argument is a string, it is interpreted as an attribute
        name, and the value of that attribute is returned for each
        element. Nested attribute names can be chained with ``.``.

        If the argument is a callable, it is called for each element,
        and the return value is included in the result. If the callable
        returns a sequence, the sequence is flattened into the result.

        Duplicate values and Nones are always filtered out.

        It is an error if a callable returns something that is not a
        model element or a flat sequence of model elements.
        """
        if isinstance(attr, str):
            attr = operator.attrgetter(attr)
        newelems: list[etree._Element] = []
        classes: set[type[GenericElement]] = set()
        for i in self:
            try:
                value = attr(i)
            except AttributeError:
                continue

            if not isinstance(value, cabc.Iterable):
                value = [value]

            for v in value:  # type: ignore[union-attr] # false-positive
                if v is None:
                    continue
                if isinstance(v, GenericElement):
                    newelems.append(v._element)
                    classes.add(type(v))
                else:
                    raise TypeError(
                        f"Map function must return a model element or a list"
                        f" of model elements, not {v!r}"
                    )

        if len(classes) == 1:
            return ElementList(self._model, newelems, classes.pop())
        return MixedElementList(self._model, newelems)


class CachedElementList(ElementList[T], t.Generic[T]):
    """An ElementList that caches the constructed proxies by UUID."""

    class _Filter(ElementList._Filter[U], t.Generic[U]):
        def __call__(
            self, *values: U, single: bool | None = None
        ) -> T | ElementList[T]:
            newlist: T | ElementList[T] = super().__call__(
                *values, single=single
            )
            if single or self._single:
                return newlist

            assert isinstance(newlist, CachedElementList)
            newlist.cacheattr = t.cast(t.Optional[str], self._parent.cacheattr)
            return newlist

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[T],
        *,
        cacheattr: str | None = None,
        **kw: t.Any,
    ) -> None:
        """Create a CachedElementList.

        Parameters
        ----------
        model
            The model that all elements are a part of.
        elements
            The members of this list.
        elemclass
            The :class:`GenericElement` subclass to use for
            reconstructing elements.
        cacheattr
            The attribute on the ``model`` to use as cache.
        """
        super().__init__(model, elements, elemclass, **kw)
        self.cacheattr = cacheattr

    def __getitem__(self, key):
        elem = super().__getitem__(key)
        if self.cacheattr and not isinstance(elem, ElementList):
            try:
                cache = getattr(self._model, self.cacheattr)
            except AttributeError:
                cache = {}
                setattr(self._model, self.cacheattr, cache)
            elem = cache.setdefault(elem.uuid, elem)
        return elem


class MixedElementList(ElementList[GenericElement]):
    """ElementList that handles proxies using ``XTYPE_HANDLERS``."""

    class _LowercaseFilter(ElementList._Filter[U], t.Generic[U]):
        def extract_key(self, element: ModelObject) -> U | str:
            value = super().extract_key(element)
            assert isinstance(value, str)
            return value.lower()

        def make_values_container(self, *values: U) -> cabc.Container[U]:
            return tuple(map(operator.methodcaller("lower"), values))

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: t.Any = None,
        **kw: t.Any,
    ) -> None:
        """Create a MixedElementList.

        Parameters
        ----------
        model
            The model that all elements are a part of.
        elements
            The members of this list.
        elemclass
            Ignored; provided for drop-in compatibility.
        """
        del elemclass
        super().__init__(model, elements, GenericElement, **kw)

    def __getattr__(
        self, attr: str
    ) -> cabc.Callable[..., GenericElement | ElementList[GenericElement]]:
        if attr == "by_type":
            return self._LowercaseFilter(self, "__class__.__name__")
        return super().__getattr__(attr)

    def __dir__(self) -> list[str]:  # pragma: no cover
        return super().__dir__() + ["by_type", "exclude_types"]


class ElementListMapKeyView(cabc.Sequence):
    def __init__(self, parent, /) -> None:
        self.__parent = parent

    @t.overload
    def __getitem__(self, idx: int) -> t.Any: ...

    @t.overload
    def __getitem__(self, idx: slice) -> list: ...

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [self.__parent._mapkey(i) for i in self.__parent[idx]]
        return self.__parent._mapkey(self.__parent[idx])

    def __len__(self) -> int:
        return len(self.__parent)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)!r})"


class ElementListMapItemsView(t.Sequence[t.Tuple[t.Any, t.Any]], t.Generic[T]):
    def __init__(self, parent, /) -> None:
        self.__parent = parent

    @t.overload
    def __getitem__(self, idx: int) -> tuple[t.Any, T]: ...

    @t.overload
    def __getitem__(self, idx: slice) -> list[tuple[t.Any, T]]: ...

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [
                (self.__parent._mapkey(i), self.__parent._map_getvalue(i))
                for i in self.__parent[idx]
            ]
        obj = self.__parent[idx]
        return (self.__parent._mapkey(obj), self.__parent._map_getvalue(obj))

    def __len__(self) -> int:
        return len(self.__parent)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)!r})"



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Custom extensions to the svgwrite ``Drawing`` object."""
from __future__ import annotations

import collections.abc as cabc
import copy
import dataclasses
import logging
import os
import re
import typing as t

from svgwrite import base, container, drawing, shapes
from svgwrite import text as svgtext

from capellambse import diagram
from capellambse import helpers as chelpers
from capellambse.diagram import capstyle

from . import decorations, generate, helpers, style, symbols

LOGGER = logging.getLogger(__name__)
DEBUG = "CAPELLAMBSE_SVG_DEBUG" in os.environ
"""Debug flag to render helping lines."""
LABEL_ICON_PADDING = 2
"""Default padding between a label's icon and text."""
NUMBER = r"-?\d+(\.\d+)?"
RE_ROTATION = re.compile(
    rf"rotate\((?P<angle>{NUMBER}),\s*(?P<x>{NUMBER}),\s*(?P<y>{NUMBER})\) "
    f"(?P<tspan_y>{NUMBER})"
)


LabelDict = t.TypedDict(
    "LabelDict",
    {
        "x": float,
        "y": float,
        "width": float,
        "height": float,
        "text": str,
        "class": str,
    },
    total=False,
)


class Drawing:
    """The main container that stores all svg elements."""

    def __init__(self, metadata: generate.DiagramMetadata):
        superparams = {
            "cursor": "pointer",
            "filename": f"{metadata.name}.svg",
            "font-family": "Open Sans",
            "font-size": "8px",
            "shape-rendering": "geometricPrecision",
            "size": metadata.size,
            "viewBox": metadata.viewbox,
        }
        if metadata.class_:
            superparams["class_"] = re.sub(r"\s+", "", metadata.class_)

        self.__drawing = drawing.Drawing(**superparams)
        self.diagram_class = metadata.class_
        self.deco_cache: set[str] = set()
        self.add_backdrop(pos=metadata.pos, size=metadata.size)

        self.obj_cache: dict[str | None, t.Any] = {}

    @property
    def filename(self) -> str:
        """Return the filename of the SVG."""
        return self.__drawing.filename

    @filename.setter
    def filename(self, name: str) -> None:
        self.__drawing.filename = name

    def save_as(self, filename: str | None = None, **kw: t.Any) -> None:
        """Write the SVG to a file.

        If ``filename`` wasn't given the underlying ``filename`` is
        taken.
        """
        kw["filename"] = filename or self.__drawing.filename
        return self.__drawing.saveas(**kw)

    def to_string(self) -> str:
        """Return a string representation of the SVG."""
        return self.__drawing.tostring()

    def add_backdrop(
        self, pos: tuple[float, float], size: tuple[float, float]
    ) -> None:
        """Add a white background rectangle to the drawing."""
        self.__backdrop = self.__drawing.rect(
            insert=pos,
            size=size,
            fill="#fff",
            stroke="none",
        )
        self.__drawing.add(self.__backdrop)

    def __repr__(self) -> str:
        return self.__drawing._repr_svg_()

    def add_rect(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        rectstyle: cabc.Mapping[str, style.Styling],
        *,
        class_: str = "",
        context_: cabc.Sequence[str] = (),
        label: LabelDict | None = None,
        description: str | None = None,
        features: cabc.Sequence[str] = (),
        id_: str | None = None,
        children: bool = False,
    ) -> container.Group:
        """Add a rectangle with auto-aligned text to the group."""
        gcls = "".join(f" context-{i}" for i in context_)
        grp: container.Group = self.__drawing.g(
            class_=f"Box {class_}{gcls}", id_=id_
        )
        text_style = rectstyle["text_style"]
        rect_style = rectstyle["obj_style"]
        transform = chelpers.get_transformation(class_, pos, size)
        rectparams: dict[str, t.Any] = {
            "insert": pos,
            "size": size,
            "class_": class_,
            **transform,
        }

        rect: shapes.Rect = self.__drawing.rect(
            **rectparams, **rect_style._to_dict()
        )
        grp.add(rect)

        if description is not None and label is not None:
            new_label: LabelDict = copy.deepcopy(label)
            new_label["text"] = description
            self._draw_label(
                LabelBuilder(
                    new_label,
                    grp,
                    labelstyle=text_style,
                    class_=class_,
                    text_anchor="middle",
                    y_margin=None,
                    icon=False,
                )
            )

        if features or class_ in decorations.needs_feature_line:
            self._draw_feature_line(rect, grp, rect_style)
            if features:
                self._draw_feature_text(
                    rect, features, class_, grp, text_style
                )

        if label:
            text_anchor = (
                "start" if class_ in decorations.start_aligned else "middle"
            )
            y_margin = None
            if children or class_ in decorations.always_top_label:
                y_margin = 5

            self._draw_label(
                LabelBuilder(
                    label,
                    grp,
                    labelstyle=text_style,
                    class_=class_,
                    text_anchor=text_anchor,
                    y_margin=y_margin,
                )
            )
        elif class_ in decorations.only_icons:
            icon_size = 50
            labelpos = (
                pos[0] + (size[0] - icon_size / 4) / 2,
                pos[1] + (decorations.feature_space - 15) / 2,
            )
            label = {
                "x": labelpos[0],
                "y": labelpos[1],
                "width": 0.0,
                "height": 0.0,
                "text": "",
                "class": "",
            }
            self.add_label_image(
                LabelBuilder(
                    label, grp, text_style, class_=class_, icon_size=icon_size
                ),
                labelpos,
            )

        if DEBUG:
            self._draw_rect_helping_lines(grp, pos, size)

        return grp

    def _draw_feature_line(
        self,
        obj: base.BaseElement,
        group: container.Group | None,
        objstyle: style.Styling | None,
    ) -> shapes.Line | None:
        """Draw a Line on the given object."""
        x, y = obj.attribs["x"], obj.attribs["y"]
        w = obj.attribs["width"]

        line = self.__drawing.line(
            start=(x, y + decorations.feature_space),
            end=(x + w, y + decorations.feature_space),
            stroke=getattr(objstyle, "stroke", None),
        )
        if group is None:
            return line

        return group.add(line)

    def _draw_feature_text(
        self,
        obj: base.BaseElement,
        features: cabc.Sequence[str],
        class_: str,
        group: container.Group,
        labelstyle: style.Styling,
    ) -> None:
        """Draw features text on given object."""
        x, y = obj.attribs["x"], obj.attribs["y"]
        w, h = obj.attribs["width"], obj.attribs["height"]
        label: LabelDict = {
            "x": x + decorations.feature_space / 2,
            "y": y + decorations.feature_space,
            "width": w - decorations.feature_space / 2,
            "height": h - decorations.feature_space,
            "class": "Features",
            "text": "\n".join(
                chelpers.flatten_html_string(feat) for feat in features
            ),
        }
        self._draw_label(
            LabelBuilder(
                label,
                group,
                class_=class_,
                labelstyle=labelstyle,
                y_margin=7,
                icon=False,
                alignment="left",
            )
        )

    def _draw_label(self, builder: LabelBuilder) -> None:
        """Draw label text on given object and return the label's group."""
        assert isinstance(builder.label["x"], (int, float))
        assert isinstance(builder.label["y"], (int, float))
        assert isinstance(builder.label["width"], (int, float))
        assert isinstance(builder.label["height"], (int, float))
        assert isinstance(builder.label.get("text", ""), str)

        text = self._make_text(builder)
        builder.icon &= f"{builder.class_}Symbol" in decorations.deco_factories
        if not text.elements:
            lines = render_hbounded_lines(builder, builder.icon)
            x, y = get_label_position(builder, lines)
            for line in lines.lines:
                params = {
                    "insert": (x, y),
                    "text": line,
                    "xml:space": "preserve",
                }
                text.add(svgtext.TSpan(**params))
                y += lines.line_height

            if builder.icon:
                icon_pos = get_label_icon_position(builder, lines)
                self.add_label_image(builder, icon_pos)
        return builder.group

    def _make_text(self, builder: LabelBuilder) -> svgtext.Text:
        """Return a text element and add it to the builder group."""
        textattrs = {
            "text": "",
            "insert": (builder.label["x"], builder.label["y"]),
            "text_anchor": builder.text_anchor,
            "dominant_baseline": "middle",
        }
        if "class" in builder.label:
            textattrs["class_"] = builder.label["class"]

        if transform := getattr(builder.labelstyle, "transform", None):
            if match := re.search(RE_ROTATION, transform):
                if match.group("angle") != "-90":
                    raise NotImplementedError(
                        "Angles other than -90 are not supported"
                    )

                x = float(match.group("x"))
                y = float(match.group("y"))
                if new_y := match.group("tspan_y"):
                    y = float(new_y)
                    transform = transform.replace(f" {new_y}", "")

            textattrs["transform"] = transform
            delattr(builder.labelstyle, "transform")

        text = self.__drawing.text(
            **textattrs, **builder.labelstyle._to_dict()
        )
        if transform:
            params = {
                "insert": (x, y),
                "text": str(builder.label["text"]),
                "xml:space": "preserve",
            }
            text.add(svgtext.TSpan(**params))
        builder.group.add(text)
        return text

    def add_label_image(
        self, builder: LabelBuilder, pos: tuple[float, float]
    ) -> None:
        """Add label svgobject to given group."""
        if builder.class_ is None:
            builder.class_ = "Error"

        if builder.class_ not in self.deco_cache:
            self._add_decofactory(builder.class_)

        builder.group.add(
            self.__drawing.use(
                href=f"#{builder.class_}Symbol",
                insert=pos,
                size=(builder.icon_size, builder.icon_size),
            )
        )

    def get_port_transformation(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        class_: str,
        parent_id: str | None,
    ) -> str:
        """Return rotation transformation for port-object-styling."""
        rect = self.obj_cache[parent_id]
        x, y = pos
        w, h = size
        angle = 0
        par_x, par_y = rect["x"] + 0.5, rect["y"] + 0.5
        if x <= par_x:
            angle = 90
        elif par_x + rect["width"] < x + w:
            angle = -90
        elif y <= par_y:
            angle = -180

        angle += 180 * (class_ in ["FOP", "CP_IN"])
        return f"rotate({angle} {x + w / 2} {y + h / 2})"

    def add_port(
        self,
        pos: tuple[float, float],
        size: tuple[float, float],
        text_style: style.Styling,
        parent_id: str | None,
        *,
        class_: str,
        context_: cabc.Sequence[str] = (),
        obj_style: style.Styling,
        label: LabelDict | None = None,
        id_: str | None = None,
    ) -> container.Group:
        gcls = "".join(f" context-{i}" for i in context_)
        grp = self.__drawing.g(class_=f"Box {class_}{gcls}", id_=id_)
        if class_ in decorations.all_directed_ports:
            port_id = "Error"
            if class_ in decorations.function_ports:
                port_id = "Port"
            elif class_ in decorations.component_ports:
                port_id = "ComponentPort"

            if port_id not in self.deco_cache:
                self._add_decofactory(port_id)

            grp.add(
                self.__drawing.use(
                    href=f"#{port_id}Symbol",
                    insert=pos,
                    size=size,
                    transform=self.get_port_transformation(
                        pos, size, class_, parent_id
                    ),
                    class_=class_,
                    **obj_style._to_dict(),
                )
            )
        else:
            grp.add(
                self.__drawing.rect(
                    insert=pos,
                    size=size,
                    class_=class_,
                    transform=self.get_port_transformation(
                        pos, size, class_, parent_id
                    ),
                    **obj_style._to_dict(),
                )
            )

        if label is not None:
            self._draw_label(
                LabelBuilder(
                    label,
                    grp,
                    class_="Annotation",
                    labelstyle=text_style,
                    text_anchor="middle",
                    y_margin=0,
                    icon=False,
                )
            )

        return grp

    def draw_object(self, obj: cabc.Mapping[str, t.Any]) -> None:
        """Draw an object into this drawing.

        Parameters
        ----------
        obj
            The (decoded) JSON-dict of a single diagram object.
        """
        mobj = copy.deepcopy(obj)
        del obj
        type_mapping: dict[str, tuple[t.Any, str]] = {
            "box": (self._draw_box, "Box"),
            "edge": (self._draw_edge, "Edge"),
            "circle": (self._draw_circle, "Edge"),
            "symbol": (self._draw_symbol, "Box"),
            "box_symbol": (self._draw_box_symbol, "Box"),
        }

        try:
            drawfunc, style_type = type_mapping[mobj["type"]]
        except KeyError:
            raise ValueError(f"Invalid object type: {mobj['type']}") from None

        objparams = {
            f"{k}_": v for k, v in mobj.items() if k not in {"type", "style"}
        }
        if mobj["class"] in decorations.all_ports:
            mobj["type"] = "box"  # type: ignore[index]

        class_: str = style_type + (
            f".{mobj['class']}" if "class" in mobj else ""
        )

        my_styles: dict[str, t.Any] = {
            **capstyle.get_style(self.diagram_class, class_),
            **mobj.get("style", {}),
        }
        obj_style = style.Styling(
            self.diagram_class,
            class_,
            **{k: v for k, v in my_styles.items() if "_" not in k},
        )
        text_style = style.Styling(
            self.diagram_class,
            class_,
            "text",
            **{
                k[5:]: v for k, v in my_styles.items() if k.startswith("text_")
            },
        )

        self.obj_cache[mobj["id"]] = mobj

        drawfunc(**objparams, obj_style=obj_style, text_style=text_style)

        self._deploy_defs(obj_style)
        self._deploy_defs(text_style)

    def _deploy_defs(self, styling: style.Styling) -> None:
        defs_ids = {d.attribs.get("id") for d in self.__drawing.defs.elements}
        for attr in styling:
            val = getattr(styling, attr)
            if isinstance(val, cabc.Iterable) and not isinstance(
                val, (str, diagram.RGB)
            ):
                grad_id = styling._generate_id("CustomGradient", val)
                if grad_id not in defs_ids:
                    gradient = symbols._make_lgradient(
                        id_=grad_id, stop_colors=val
                    )
                    self.__drawing.defs.add(gradient)
                    defs_ids.add(grad_id)

        defaultstyles = diagram.get_style(self.diagram_class, styling._class)

        def getstyleattr(sobj: object, attr: str) -> t.Any:
            return getattr(sobj, attr, None) or defaultstyles.get(
                styling._style_name(attr)
            )

        markers = (
            getstyleattr(super(), "marker-start"),
            getstyleattr(super(), "marker-end"),
        )
        for marker in markers:
            if marker is None:
                continue

            stroke = str(getstyleattr(styling, "stroke"))
            stroke_width = str(getstyleattr(styling, "stroke-width"))
            marker_id = styling._generate_id(marker, [stroke])
            if marker_id not in defs_ids:
                factory = decorations.deco_factories[marker]
                markstyle = style.Styling(
                    self.diagram_class,
                    styling._class,
                    _prefix=styling._prefix,
                    _markers=False,
                    stroke=stroke,
                    stroke_width=stroke_width,
                )._to_dict()
                markstyle.pop("marker-start", None)
                markstyle.pop("marker-end", None)
                self.__drawing.defs.add(
                    factory.function(marker_id, **markstyle)
                )
                defs_ids.add(marker_id)

    def _draw_symbol(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        context_: cabc.Sequence[str] = (),
        parent_: str | None = None,
        label_: LabelDict | None = None,
        id_: str | None = None,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ):
        del kw  # Dismiss additional info from json
        assert isinstance(label_, (dict, type(None)))
        pos = (x_ + 0.5, y_ + 0.5)
        size = (width_, height_)
        if (
            class_ not in self.deco_cache
            and class_ not in decorations.all_ports
        ):
            self._add_decofactory(class_)

        if class_ in decorations.all_ports:
            grp = self.add_port(
                pos,
                size,
                text_style,
                parent_,
                class_=class_,
                context_=context_,
                obj_style=obj_style,
                label=label_,
                id_=id_,
            )
        else:
            gcls = "".join(f" context-{i}" for i in context_)
            grp = self.__drawing.g(class_=f"Box {class_}{gcls}", id_=id_)
            grp.add(
                self.__drawing.use(
                    href=f"#{class_}Symbol",
                    insert=pos,
                    size=size,
                    class_=class_,
                    **obj_style._to_dict(),
                )
            )

        self.__drawing.add(grp)
        if label_:
            label_["class"] = "Annotation"
            self._draw_label(
                LabelBuilder(
                    label_,
                    grp or self.__drawing.elements[-1],
                    labelstyle=text_style,
                    class_=class_,
                    y_margin=0,
                    text_anchor="middle",
                    icon=False,
                )
            )

    def _draw_box(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        context_: cabc.Sequence[str] = (),
        children_: cabc.Sequence[str] = (),
        features_: cabc.Sequence[str] = (),
        label_: str | LabelDict | None = None,
        description_: str | None = None,
        id_: str,
        class_: str,
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ) -> container.Group:
        del kw  # Dismiss additional info from json, for e.g.: ports_
        pos = (x_ + 0.5, y_ + 0.5)
        size = (width_, height_)
        rect_style = {"text_style": text_style, "obj_style": obj_style}
        label: LabelDict | None = None
        if isinstance(label_, str):
            label = {
                "x": x_,
                "y": y_,
                "width": width_,
                "height": height_,
                "text": label_,
                "class": class_ or "Box",
            }
        elif label_ is not None:
            label = label_
            label["class"] = "Annotation"

        grp = self.add_rect(
            pos,
            size,
            rect_style,
            class_=class_,
            context_=context_,
            id_=id_,
            label=label,
            description=description_,
            features=features_,
            children=bool(children_),
        )
        return self.__drawing.add(grp)

    def _draw_box_symbol(
        self,
        *,
        x_: int | float,
        y_: int | float,
        width_: int | float,
        height_: int | float,
        label_: LabelDict,
        id_: str,
        class_: str,
        context_: cabc.Sequence[str] = (),
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw: t.Any,
    ) -> container.Group:
        del kw

        grp = self._draw_box(
            x_=x_,
            y_=y_,
            width_=width_,
            height_=height_,
            id_=id_,
            class_=class_,
            context_=context_,
            obj_style=obj_style,
            text_style=text_style,
        )
        label_["class"] = "Annotation"
        self._draw_label(
            LabelBuilder(
                label_,
                grp or self.__drawing.elements[-1],
                labelstyle=text_style,
                class_=class_,
                y_margin=0,
                text_anchor="middle",
                icon=False,
            )
        )
        return grp

    def _draw_circle(
        self,
        *,
        center_: tuple[float, float],
        radius_: float,
        id_: str | None = None,
        class_: str | None = None,
        context_: cabc.Sequence[str] = (),
        obj_style: style.Styling,
        text_style: style.Styling,
    ) -> container.Group:
        del text_style  # No label for circles
        center_ = (center_[0] + 0.5, center_[1] + 0.5)
        obj_style.fill = obj_style.stroke or diagram.RGB(0, 0, 0)
        del obj_style.stroke
        gcls = "".join(f" context-{i}" for i in context_)
        grp = self.__drawing.g(class_=f"Circle {class_}{gcls}", id_=id_)
        circle = self.__drawing.circle(
            center=center_, r=radius_, **obj_style._to_dict()
        )
        grp.add(circle)
        return self.__drawing.add(grp)

    def _draw_edge(
        self,
        *,
        points_: list[list[int]],
        labels_: t.Sequence[LabelDict] = (),
        id_: str,
        class_: str,
        context_: cabc.Sequence[str] = (),
        obj_style: style.Styling,
        text_style: style.Styling,
        **kw,
    ) -> container.Group:
        del kw
        points: list = [(x + 0.5, y + 0.5) for x, y in points_]
        gcls = "".join(f" context-{i}" for i in context_)
        grp = self.__drawing.g(class_=f"Edge {class_}{gcls}", id_=id_)
        path = self.__drawing.path(
            d=["M"] + points, class_="Edge", **obj_style._to_dict()
        )
        grp.add(path)

        # Received text space doesn't allow for anything else than the text
        for label_ in labels_:
            label_["class"] = "Annotation"
            self._draw_edge_label(
                label_,
                grp,
                labelstyle=text_style,
                text_anchor="middle",
                class_=class_,
                y_margin=0,
            )
        return self.__drawing.add(grp)

    def _draw_edge_label(
        self,
        label: LabelDict,
        group: container.Group,
        *,
        class_: str,
        labelstyle: style.Styling,
        text_anchor: str = "start",
        y_margin: int | float,
    ) -> container.Group:
        if f"{class_}Symbol" in decorations.deco_factories:
            additional_space = (
                decorations.icon_size + 2 * decorations.icon_padding
            )
            label["width"] += additional_space + 2
            label["x"] -= additional_space / 2

        self._draw_label(
            LabelBuilder(
                label,
                group,
                labelstyle=labelstyle,
                class_=class_,
                y_margin=y_margin,
                text_anchor=text_anchor,
            )
        )

        return group

    def _draw_line(
        self,
        data: dict[str, float | int],
        *,
        obj_style: style.Styling,
    ) -> shapes.Line | None:
        """Draw a Line on the given object."""
        x1, y1 = data["x"], data["y"]
        x2 = data.get("x1") or data["x"] + data["width"]
        y2 = data.get("y1") or data["y"]
        line = self.__drawing.line(
            start=(x1, y1), end=(x2, y2), **obj_style._to_dict()
        )
        return line

    def _draw_rect_helping_lines(
        self,
        grp: container.Group,
        pos: tuple[float, float],
        size: tuple[float, float],
    ) -> None:
        linestyle = style.Styling(
            self.diagram_class,
            "Edge",
            stroke="rgb(239, 41, 41)",
            **{"stroke-dasharray": "5"},
        )
        start = (pos[0] + size[0] / 2, pos[1])
        end = (pos[0] + size[0] / 2, pos[1] + size[1])
        ln = self.__drawing.line(start=start, end=end, **linestyle._to_dict())
        grp.add(ln)

        y = pos[1] + size[1] / 2
        start = (pos[0], y)
        end = (pos[0] + size[0], y)
        ln = self.__drawing.line(start=start, end=end, **linestyle._to_dict())
        grp.add(ln)

    def _add_decofactory(self, name: str) -> None:
        factory = decorations.deco_factories[f"{name}Symbol"]
        self.__drawing.defs.add(factory.function())
        for dep in factory.dependencies:
            if dep not in self.deco_cache:
                self._add_decofactory(dep)

        self.deco_cache.add(name)


@dataclasses.dataclass
class LabelBuilder:
    """Helper data-class for building labels."""

    label: LabelDict
    group: container.Group
    labelstyle: style.Styling
    class_: str | None = None
    y_margin: int | float | None = None
    text_anchor: str = "start"
    icon: bool = True
    icon_size: float | int = decorations.icon_size
    alignment: helpers.AlignmentLiteral = "center"


class LinesData(t.NamedTuple):
    """Helper data-tuple for rendering text-lines from labels."""

    lines: list[str]
    line_height: float
    text_height: float
    margin: float
    max_line_width: float


def render_hbounded_lines(
    builder: LabelBuilder, render_icon: bool
) -> LinesData:
    """Return Lines data to render a label."""
    (
        lines,
        label_margin,
        max_text_width,
    ) = helpers.check_for_horizontal_overflow(
        str(builder.label["text"]),
        builder.label["width"],
        decorations.icon_padding if render_icon else 0,
        builder.icon_size if render_icon else 0,
        builder.alignment,
    )
    lines_to_render = helpers.check_for_vertical_overflow(
        lines, builder.label["height"], max_text_width
    )
    line_height = max(j for _, j in map(chelpers.extent_func, lines_to_render))
    text_height = line_height * len(lines_to_render)
    max_line_width = max(
        w for w, _ in map(chelpers.extent_func, lines_to_render)
    )
    assert max_text_width >= max_line_width
    return LinesData(
        lines_to_render, line_height, text_height, label_margin, max_line_width
    )


def get_label_position(
    builder: LabelBuilder, lines: LinesData
) -> tuple[float, float]:
    """Return the initial label position (x, y)."""
    icon_size = (builder.icon_size + decorations.icon_padding) * builder.icon
    if builder.text_anchor == "start":
        builder.label["x"] += lines.margin + icon_size
    else:
        builder.label["x"] += (builder.label["width"] + icon_size) / 2

    dominant_baseline_adjust = chelpers.extent_func(lines.lines[0])[1] / 2
    if builder.y_margin is None:
        builder.y_margin = (builder.label["height"] - lines.text_height) / 2
    builder.label["y"] += builder.y_margin + dominant_baseline_adjust
    return builder.label["x"], builder.label["y"]


def get_label_icon_position(
    builder: LabelBuilder, lines: LinesData
) -> tuple[float, float]:
    """Calculate the icon's position."""
    icon_y = builder.label["y"] - builder.icon_size / 2
    if lines.text_height > lines.line_height:
        icon_y += lines.text_height / 2 - decorations.icon_padding * 2
        if builder.label["class"] == "Annotation":
            icon_y -= decorations.icon_padding

    icon_x = (
        builder.label["x"] - builder.icon_size - decorations.icon_padding * 2
    )
    if builder.text_anchor == "middle":
        icon_x -= lines.max_line_width / 3
    return (icon_x, icon_y)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Utility classes that are used across all layers.

.. diagram:: [CDB] BaseLayer
"""
from .. import common as c
from . import cs, information


class BaseArchitectureLayer(c.GenericElement):
    """A template architecture layer."""

    _xmltag = "ownedArchitectures"

    data_package = c.DirectProxyAccessor(information.DataPkg)
    interface_package = c.DirectProxyAccessor(cs.InterfacePkg)

    all_classes = c.DeepProxyAccessor(information.Class, aslist=c.ElementList)
    all_collections = c.DeepProxyAccessor(
        information.Collection, aslist=c.ElementList
    )
    all_unions = c.DeepProxyAccessor(information.Union, aslist=c.ElementList)
    all_enumerations = c.DeepProxyAccessor(
        information.datatype.Enumeration, aslist=c.ElementList
    )
    all_complex_values = c.DeepProxyAccessor(
        information.datavalue.ComplexValue, aslist=c.ElementList
    )
    all_interfaces = c.DeepProxyAccessor(cs.Interface, aslist=c.ElementList)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for handling ReqIF Requirements.

.. diagram:: [CDB] Requirements ORM
"""
import typing as t

from ._capellareq import *
from ._glue import *
from ._requirements import *

if not t.TYPE_CHECKING:
    _deprecated_names = {
        "EnumDataTypeDefinition": EnumerationDataTypeDefinition,
        "RequirementsFolder": Folder,
        "RequirementsIncRelation": CapellaIncomingRelation,
        "RequirementsIntRelation": InternalRelation,
        "RequirementsModule": CapellaModule,
        "RequirementsOutRelation": CapellaOutgoingRelation,
        "RequirementsTypesFolder": CapellaTypesFolder,
    }

    def __getattr__(name) -> type[ReqIFElement]:
        import warnings

        if cls := _deprecated_names.get(name):
            warnings.warn(
                f"Name {name} is deprecated, use {cls.__name__} instead",
                DeprecationWarning,
                stacklevel=2,
            )

            return cls

        elif name.startswith("XT_"):
            warnings.warn(
                (
                    "XT_* strings are deprecated, use the respective classes"
                    " and 'build_xtype(cls)' instead"
                ),
                DeprecationWarning,
                stacklevel=2,
            )

            from . import elements

            return getattr(elements, name)

        elif name == "elements":
            import importlib

            fullname = f"{__name__}.{name}"
            warnings.warn(
                f"{fullname!r} is deprecated, use {__name__!r} instead",
                DeprecationWarning,
                stacklevel=2,
            )

            return importlib.import_module(fullname)

        else:
            raise NameError(f"No name {name} in module {__name__}")

    __all__: list[str] = ["elements"]
    from ._capellareq import __all__ as _cr_all
    from ._requirements import __all__ as _rq_all
    from .elements import __all__ as _el_all

    __all__ += _cr_all + _rq_all + _el_all
    del _cr_all, _rq_all, _el_all
del t
